from typing import Iterable, List, Optional
from spacy.language import Language
from spacy.tokens import Doc
from jange.base import Operation
from jange.stream import DataStream
from jange import config
from .utils import cached_spacy_model


class SpacyBasedOperation(Operation):
    def __init__(self, nlp: Optional[Language] = None, name: str = "spacy_op") -> None:
        super().__init__(name=name)
        self.nlp = nlp or cached_spacy_model(config.DEFAULT_SPACY_MODEL)
        self.model_path = self.nlp.path

    def get_docs(self, ds: DataStream) -> Iterable[Doc]:
        """Returns an interable of spacy Doc from the datastream.
        If the data stream already contains spacy Docs then they
        are returned as-is otherwise the nlp object is used to
        create spacy Docs

        Parameters
        ----------
        ds : DataStream
            input data stream

        Returns
        ------
        out : Iterable[spacy.tokens.Doc]
            an iterable of spacy's `Doc` objects
        """
        if ds.item_type != Doc:
            docs = self.nlp.pipe(ds)
        else:
            docs = ds

        return docs

    def discard_tokens_from_doc(self, doc: Doc, token_ids: List[int]) -> Doc:
        """Returns a new document after discarding the tokens

        Parameters
        ----------
        doc : spacy.tokens.Doc
            orignal document
        token_ids : List[int]
            a list of index of tokens to discard

        Returns
        -------
        out : spacy.tokens.Doc
            a new document which does not contain the tokens specified
        """
        tokens = [t for t in doc if t.i not in token_ids]
        words = [t.text for t in tokens]
        spaces = [t.whitespace_ == " " for t in tokens]
        spaces[-1] = False
        return Doc(self.nlp.vocab, words=words, spaces=spaces)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["nlp"]
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        nlp = cached_spacy_model(state["model_path"])
        self.nlp = nlp



from typing import Any, List, Optional, Iterable
import cytoolz
from scipy.sparse import issparse


class OperationCollection(list):
    def __add__(self, other):
        result = super().__add__(other)
        return OperationCollection(result)

    def find_by_name(self, name, first_only=True):
        if first_only:
            return next((op for op in self if op.name == name))
        else:
            return [op for op in self if op.name == name]


class DataStream:
    """A class representing a stream of data. A data stream is created as
    a result of some operation. DataStream object can be iterated which
    basically iterates through the underlying data. The underlying data
    is stored in `items` attribute which can be any iterable object.

    Parameters
    ----------
    items : iterable
        an iterable that contains the raw data

    applied_ops : Optional[List[Operation]]
        a list of operations that were applied to create this stream of data

    Example
    -------
    >>> ds = DataStream(items=[1, 2, 3])
    >>> print(list(ds))
    >>> [1, 2, 3]


    Attributes
    ----------
    applied_ops : List[Operation]
        a list of operations that were applied to create this stream of data

    items : iterable
        an iterable that contains the raw data
    """

    def __init__(
        self,
        items: Iterable[Any],
        applied_ops: Optional[List] = None,
        context: Optional[Iterable[Any]] = None,
    ):
        self._validate_items_or_raise(items)

        self.items = items
        self.context = context

        if applied_ops and not isinstance(applied_ops, OperationCollection):
            applied_ops = OperationCollection(applied_ops)
        self.applied_ops = applied_ops or OperationCollection()

    def _is_countable(self, x):
        return hasattr(x, "__len__")

    def _validate_items_or_raise(self, items):
        if items is None:
            raise ValueError("items cannot be None")

        if self._is_countable(items):
            # sparse matrix don't support calling len
            # so use getnnz()
            if issparse(items):
                count = items.getnnz()
            else:
                count = len(items)

            if count == 0:
                raise ValueError(
                    f"items must have atleast one element. received {items}"
                )

    def __iter__(self):
        for item in self.items:
            yield item

    @property
    def is_countable(self) -> bool:
        return self._is_countable(self.items)

    @property
    def item_type(self):
        if self.is_countable:
            return type(self.items[0])
        else:
            first, items = cytoolz.peek(self.items)
            self.items = items
            return type(first)

    def apply(self, *ops):
        x = self
        for op in ops:
            x = op.run(x)
        return x


class Operation:
    def __init__(self, name: Optional[str] = None) -> None:
        super().__init__()
        name = name or self.__class__.__name__
        self.name = name

    def run(self, ds: DataStream) -> DataStream:
        raise NotImplementedError()

    def __call__(self, ds: DataStream) -> DataStream:
        return self.run(ds)


class TrainableMixin:
    def __init__(self) -> None:
        self.should_train = True



from typing import Optional
import sklearn.cluster as skcluster
from sklearn.base import ClusterMixin

from jange.base import Operation, TrainableMixin
from jange.stream import DataStream

# some algorithms do not support predicting on new samples
# and needs retraining all the time. separate those algorithms
ALGORITHMS_SUPPORTING_NEW_INFERENCE = [
    skcluster.KMeans,
    skcluster.MiniBatchKMeans,
    skcluster.AffinityPropagation,
    skcluster.Birch,
    skcluster.MeanShift,
]
ALGORITHMS_NOT_SUPPORTING_NEW_INFERENCE = [
    skcluster.AgglomerativeClustering,
    skcluster.DBSCAN,
    skcluster.OPTICS,
    skcluster.SpectralClustering,
]

SUPPORTED_CLASSES = (
    ALGORITHMS_SUPPORTING_NEW_INFERENCE + ALGORITHMS_NOT_SUPPORTING_NEW_INFERENCE
)


class ClusterOperation(Operation, TrainableMixin):
    def __init__(self, model: ClusterMixin, name: Optional[str] = "cluster") -> None:
        super().__init__(name=name)
        if not any(isinstance(model, cls) for cls in SUPPORTED_CLASSES):
            raise ValueError(
                f"model should be one of {SUPPORTED_CLASSES} but got {type(model)}"
            )

        self.model: ClusterMixin = model

    def run(self, ds: DataStream) -> DataStream:
        vectors = ds.items

        # if model does not have predict method then
        # regardless of training state, call fit_predict
        if not hasattr(self.model, "predict"):
            clusters = self.model.fit_predict(vectors)
        else:
            if self.should_train:
                self.model.fit(vectors)

            clusters = self.model.predict(vectors)
        return DataStream(
            clusters, applied_ops=ds.applied_ops + [self], context=ds.context
        )


def kmeans(n_clusters: int, **kwargs) -> ClusterOperation:
    model = skcluster.KMeans(n_clusters=n_clusters, **kwargs)
    return ClusterOperation(model=model, name="kmeans")


def minibatch_kmeans(n_clusters: int, **kwargs) -> ClusterOperation:
    model = skcluster.MiniBatchKMeans(n_clusters=n_clusters, **kwargs)
    return ClusterOperation(model=model, name="minibatch_kmeans")



"""This module contains implementation of Topic-modeling algorithms
"""

from typing import Optional

from sklearn.decomposition import NMF, LatentDirichletAllocation
from sklearn.base import TransformerMixin

from jange.base import Operation, TrainableMixin, DataStream


class TopicModelingOperation(Operation, TrainableMixin):
    def __init__(
        self,
        model: TransformerMixin,
        feature_names: list,
        max_words_per_topic: Optional[int] = 10,
        name: Optional[str] = "topic_modeling",
    ) -> None:
        super().__init__(name=name)
        self.model: TransformerMixin = model
        self.feature_names = feature_names
        self.max_words_per_topic = max_words_per_topic
        self.topics = []

    def run(self, ds: DataStream) -> DataStream:
        if self.should_train:
            self.model.fit(ds.items)

            for i, topic_vec in enumerate(self.model.components_):
                words_of_topic = []
                for feature_id in topic_vec.argsort()[
                    -1 : -self.max_words_per_topic - 1 : -1
                ]:
                    words_of_topic.append(self.feature_names[feature_id])
                self.topics.append(words_of_topic)

        features = self.model.transform(ds.items)

        topic_per_item = []
        # features.shape = [n_items, n_topics]
        # sort by score per topic in ascending order and take
        # the last one (with the heighest score) for each row
        # using [:,-1] slice
        for topic_idx in features.argsort(axis=1)[:, -1]:
            topic_per_item.append(self.topics[topic_idx])

        return DataStream(
            topic_per_item, applied_ops=ds.applied_ops + [self], context=ds.context
        )


def nmf(
    n_topics: int,
    feature_names: list,
    max_words_per_topic: Optional[int] = 10,
    name: Optional[str] = "nmf",
) -> TopicModelingOperation:
    model = NMF(n_components=n_topics)
    return TopicModelingOperation(
        model=model,
        feature_names=feature_names,
        max_words_per_topic=max_words_per_topic,
        name=name,
    )


def lda(
    n_topics: int,
    feature_names: list,
    max_words_per_topic: Optional[str] = 10,
    name: Optional[str] = "lda",
) -> TopicModelingOperation:
    model = LatentDirichletAllocation(n_components=n_topics)
    return TopicModelingOperation(
        model=model,
        feature_names=feature_names,
        max_words_per_topic=max_words_per_topic,
        name=name,
    )



"""This module contains commonly used dimension reduction algorithms
"""
from sklearn.base import TransformerMixin
import sklearn.decomposition as skd
import sklearn.manifold as skm
from scipy.sparse import issparse

from jange.stream import DataStream
from jange.base import Operation, TrainableMixin


# some algorithms do not support predicting on new samples
# and needs retraining all the time. separate those algorithms
ALGORITHMS_SUPPORTING_NEW_INFERENCE = [
    skm.Isomap,
    skm.LocallyLinearEmbedding,
    skd.DictionaryLearning,
    skd.FactorAnalysis,
    skd.FastICA,
    skd.IncrementalPCA,
    skd.KernelPCA,
    skd.LatentDirichletAllocation,
    skd.MiniBatchDictionaryLearning,
    skd.MiniBatchSparsePCA,
    skd.NMF,
    skd.PCA,
    skd.SparsePCA,
    skd.SparseCoder,
]
ALGORITHMS_NOT_SUPPORTING_NEW_INFERENCE = [skm.MDS, skm.SpectralEmbedding, skm.TSNE]

SUPPORTED_CLASSES = (
    ALGORITHMS_SUPPORTING_NEW_INFERENCE + ALGORITHMS_NOT_SUPPORTING_NEW_INFERENCE
)


class DimensionReductionOperation(Operation, TrainableMixin):
    """Operation for reducing dimension of a multi-dimensional array.
    This operation is primarily used for reducing large feature space
    to 2D or 3D for easy visualization.

    Parameters
    ----------
    model : TransformerMixin
        a scikit-learn model that reduces the dimensions. Usually it will
        be PCA or TSNE. See `SUPPORTED_CLASSES` for all scikit-learn models
        that are supported
    name : str
        name of this operation
    """

    def __init__(self, model: TransformerMixin, name: str = "dim_reduction") -> None:
        super().__init__(name=name)
        if not any(isinstance(model, cls) for cls in SUPPORTED_CLASSES):
            raise ValueError(
                f"model should be one of {SUPPORTED_CLASSES} but got {type(model)}"
            )
        self.model: TransformerMixin = model

    def run(self, ds: DataStream) -> DataStream:
        features = ds.items
        if issparse(features):
            features = features.toarray()

        # if model has no transform method, then we fit the model
        # regardless of the state of `should_train` flag
        if not hasattr(self.model, "transform"):
            reduced_features = self.model.fit_transform(features)
        elif self.should_train:
            reduced_features = self.model.fit_transform(features)
        else:
            reduced_features = self.model.transform(features)

        return DataStream(
            reduced_features, applied_ops=ds.applied_ops + [self], context=ds.context
        )


def pca(n_dim: int = 2,):
    """DimensionReductionOperation with PCA

    Parameters
    ----------
    n_dim : int, optional
        reduce the original n-dimensional array to `n_dim` array, by default 2

    Returns
    -------
    DimensioonReductionOperation
    """
    model = skd.PCA(n_components=n_dim)
    return DimensionReductionOperation(model, name="pca")


def tsne(n_dim: int = 2):
    """DimensionReductionOperation with TSNE

    Parameters
    ----------
    n_dim : int, optional
        reduce the original n-dimensional array to `n_dim` array, by default 2

    Returns
    -------
    DimensioonReductionOperation
    """
    model = skm.TSNE(n_components=n_dim)
    return DimensionReductionOperation(model, name="tsne")



from typing import Optional, Tuple, Union

import numpy as np
from sklearn.feature_extraction.text import (
    TfidfVectorizer,
    CountVectorizer,
)
from spacy.language import Language

from jange.base import Operation, TrainableMixin
from jange.ops.base import SpacyBasedOperation
from jange.stream import DataStream


class SklearnBasedVectorizer(Operation, TrainableMixin):
    """Vectorize texts based on algorithms available in scikit-learn.
    TfidfVectorizer, CountVectorizer and HashingVectorizer can be used
    as the underlying model.

    Paramters
    ----------
    model : Union[TfidfVectorizer, CountVectorizer, HashingVectorizer]
        any of TfidfVectorizer, CountVectorizer, HashingVectorizer instance

    name : str
        name of this operation

    Example
    -------
    >>> ds = DataStream(["this is text1", "this is text2"])
    >>> op = SklearnBasedVectorizer(model=TfidfVectorizer(max_features=1000), name='tfidf')
    >>> ds.apply(op)

    Attributes
    ----------
    model : Union[TfidfVectorizer, CountVectorizer]
        underlying model for vectorizing texts

    name : str
        name of this operation
    """

    def __init__(
        self,
        model: Union[TfidfVectorizer, CountVectorizer],
        name: str = "sklearn_vectorizer",
    ) -> None:
        super().__init__(name=name)
        self.model = model

    def run(self, ds: DataStream) -> DataStream:
        x = list(ds.items)
        if not isinstance(x[0], str):
            x = list(map(str, x))
        if self.should_train:
            self.model.fit(x)

        vectors = self.model.transform(x)
        return DataStream(vectors, ds.applied_ops + [self], context=ds.context)


def tfidf(
    max_features: Optional[int] = None,
    max_df: Union[int, float] = 1.0,
    min_df: Union[int, float] = 1,
    ngram_range: Tuple[int, int] = (1, 1),
    norm: str = "l2",
    use_idf: bool = True,
    name: str = "tfidf",
    **kwargs
) -> SklearnBasedVectorizer:
    """Returns tfidf based feature vector extraction.
    Uses sklearn's TfidfVectorizer as underlying model.

    Parameters
    ----------
    max_features : Optional[int]
        If some value is provided then only top `max_features` words
        order by their count frequency are considered in the vocabulary

    max_df : Union[int, float]
        When building vocabulary, ignore terms that have document frequency higher
        than the given value. If the value is float, then it is considered as a ratio.

    min_df : Union[int, float]
        When building vocabulary, ignore terms that have document frequency less than
        the given valu. If the value is float, then it is considered as a ratio.

    ngram_range : Tuple[int, int]
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n
        will be used. For example an ``ngram_range`` of ``(1, 1)`` means only
        unigrams, ``(1, 2)`` means unigrams and bigrams, and ``(2, 2)`` means
        only bigrams.

    norm : str
        Each output row will have unit norm, either:
        * 'l2': Sum of squares of vector elements is 1. The cosine
        similarity between two vectors is their dot product when l2 norm has
        been applied.
        * 'l1': Sum of absolute values of vector elements is 1.
        See :func:`preprocessing.normalize`.

    use_idf : bool
        Enable inverse-document-frequency reweighting.

    name : str
        name of this operation

    **kwargs
        Keyword parameters that will be passed to the initializer of CountVectorizer

    See https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html
    for details on the paramters and more examples.

    Returns
    -------
    SklearnBasedVectorizer
    """
    model = TfidfVectorizer(
        max_features=max_features, max_df=max_df, min_df=min_df, **kwargs
    )
    return SklearnBasedVectorizer(model=model, name=name)


def count(
    max_features: Optional[int] = None,
    max_df: Union[int, float] = 1.0,
    min_df: Union[int, float] = 1,
    ngram_range: Tuple[int, int] = (1, 1),
    name: Optional[str] = "count",
    **kwargs
) -> SklearnBasedVectorizer:
    """Returns count based feature vector extraction.
    Uses sklearn's CountVectorizer as underlying model.

    Parameters
    ----------
    max_features : Optional[int]
        If some value is provided then only top `max_features` words
        order by their count frequency are considered in the vocabulary

    max_df : Union[int, float]
        When building vocabulary, ignore terms that have document frequency higher
        than the given value. If the value is float, then it is considered as a ratio.

    min_df : Union[int, float]
        When building vocabulary, ignore terms that have document frequency less than
        the given valu. If the value is float, then it is considered as a ratio.

    ngram_range : Tuple[int, int]
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n
        will be used. For example an ``ngram_range`` of ``(1, 1)`` means only
        unigrams, ``(1, 2)`` means unigrams and bigrams, and ``(2, 2)`` means
        only bigrams.

    name : str
        name of this operation

    **kwargs
        Keyword parameters that will be passed to the initializer of CountVectorizer

    See https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html
    for details on the paramters and more examples.

    Returns
    -------
    SklearnBasedVectorizer
    """
    model = CountVectorizer(
        max_features=max_features, max_df=max_df, min_df=min_df, **kwargs
    )
    return SklearnBasedVectorizer(model=model, name=name)


class DocumentEmbeddingOperation(SpacyBasedOperation):
    """Operation to calculate document's vector using word-embeddings.
    Word embedding of each token are collected and averaged.

    Parameters
    ----------
    nlp : Optional[Language]
        a spacy model

    name : str
        name of this operation

    Example
    -------
    >>> ds = DataStream(["this is text 1", "this is text 2"])
    >>> vector_ds = ds.apply(DocumentEmbeddingOperation())
    >>> print(vector_ds.items)

    Attributes
    ----------
    nlp : Language
        spacy model

    name : str
        name of this operation
    """

    def __init__(
        self, nlp: Optional[Language] = None, name: str = "doc_embedding"
    ) -> None:
        super().__init__(nlp=nlp, name=name)

    def run(self, ds: DataStream) -> DataStream:
        docs = self.get_docs(ds)
        vecs = []
        for d in docs:
            if d.has_vector:
                vecs.append(d.vector)
            else:
                raise Exception(
                    "A spacy model with `vectors` capability is needed for extracting document vectors"
                )

        vecs = np.array(vecs)
        return DataStream(
            items=vecs, applied_ops=ds.applied_ops + [self], context=ds.context
        )


def doc_embedding(
    nlp: Optional[Language] = None, name: str = "doc_embedding"
) -> DocumentEmbeddingOperation:
    """Helper function to return DocumentEmbeddingOperation

    Parameters
    ----------
    nlp : Optional[Language]
        a spacy model

    name : str
        name of this operation

    Returns
    -------
    DocumentEmbeddingOperation
    """
    return DocumentEmbeddingOperation(nlp=nlp, name=name)



from typing import Optional, List, Dict, Tuple

from spacy.tokens import Doc
from spacy.language import Language
from spacy.matcher import Matcher

from jange.stream import DataStream
from ..base import Operation, SpacyBasedOperation


class CaseChangeOperation(Operation):
    """Operation for changing case of the texts.

    Parameters
    ----------
    mode : str
        one of `lower`, `upper` or `capitalize`

    name : str
        name of this operation

    Example
    --------
    >>> ds = DataStream(["AAA", "Bbb"])
    >>> list(ds.apply(CaseChangeOperation(mode="lower)))
    ["aaa", "bbb"]

    Attributes
    ----------
    mode : str
        one of ['lower', 'capitalize', 'upper']

    name : str
        name of this operation
    """

    def __init__(self, mode: str = "lower", name: str = "case_change"):
        super().__init__(name=name)
        valid_modes = ["lower", "upper", "capitalize"]
        mode = mode.lower()
        if mode not in valid_modes:
            raise ValueError(
                f"Invalid value for mode passed."
                f" Expected one of {valid_modes} but received {mode}"
            )
        self.mode = mode

    def run(self, ds: DataStream):
        if self.mode == "upper":
            fn = str.upper
        elif self.mode == "capitalize":
            fn = str.capitalize
        else:
            fn = str.lower
        items = map(fn, ds)
        return DataStream(
            applied_ops=ds.applied_ops + [self], items=items, context=ds.context
        )

    def __repr__(self):
        return f"CaseChangeOperation(mode='{self.mode}')"


def lowercase(name="lowercase") -> CaseChangeOperation:
    """Helper function to create CaseChangeOperation with mode="lower"
    """
    return CaseChangeOperation(mode="lower", name=name)


def uppercase(name="uppercase") -> CaseChangeOperation:
    """Helper function to create CaseChangeOperation with mode="upper"
    """
    return CaseChangeOperation(mode="upper", name=name)


class ConvertToSpacyDocOperation(SpacyBasedOperation):
    """Convert a stream of texts to stream of spacy's `Doc`s.
    Once spacy processes a text, it creates an instance of `Doc`
    which contains a lot of information like part of speech, named
    entities and many others. It is usually better to convert texts
    to spacy's Doc and perform operations on them. For example, spacy
    has powerful pattern matching features which can be used.

    Any operation that expects a `nlp` object can benefit if you pass
    a stream of spacy `Doc`s instead of stream of strings. Otherwise those
    operations will independently convert the raw texts into spacy `Doc`
    everytime you call them!

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Example
    -------
    >>> ds = DataStream(["this is text 1", "this is text 2"])
    >>> op = ConvertToSpacyDocOperation(nlp=nlp)
    >>> ds.apply(op)

    Attributes
    ---------
    nlp : spacy.language.Language
        spacy's language model

    name : str
        name of this operation
    """

    def __init__(
        self,
        nlp: Optional[Language] = None,
        name: Optional[str] = "convert_to_spacy_doc",
    ) -> None:
        super().__init__(nlp, name=name)

    def run(self, ds: DataStream) -> DataStream:
        docs = self.get_docs(ds)
        return DataStream(docs, applied_ops=ds.applied_ops + [self], context=ds.context)


def convert_to_spacy_doc(
    nlp: Optional[Language] = None, name: str = "convert_to_spacy_doc"
) -> ConvertToSpacyDocOperation:
    """Helper function to return ConvertToSpacyDocOperation

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    out : ConvertToSpacyDocOperation
    """
    return ConvertToSpacyDocOperation(nlp=nlp, name=name)


class LemmatizeOperation(SpacyBasedOperation):
    """Perform lemmatization using spacy's language model

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Example
    -------
    >>> nlp = spacy.load("en_core_web_sm")
    >>> op = LemmatizeOperation(nlp=nlp)
    >>> ds = DataStream(["oranges are good"])
    >>> print(list(ds.apply(op))
    ["orange be good"]

    Attributes
    ---------
    nlp : spacy.language.Language
        spacy's language model

    name : str
        name of this operation
    """

    def __init__(
        self, nlp: Optional[Language] = None, name: Optional[str] = "lemmatize"
    ) -> None:
        super().__init__(nlp, name=name)

    def _get_lemmatized_doc(self, doc):
        lemma_tokens = [t.lemma_ for t in doc]
        return self.nlp.make_doc(" ".join(lemma_tokens))

    def run(self, ds: DataStream):
        docs = self.get_docs(ds)
        items = map(self._get_lemmatized_doc, docs)
        for _, proc in self.nlp.pipeline:
            items = proc.pipe(items)
        return DataStream(
            applied_ops=ds.applied_ops + [self], items=items, context=ds.context
        )

    def __repr__(self):
        return "LemmatizeOperation()"


def lemmatize(nlp: Optional[Language] = None, name="lemmatize") -> LemmatizeOperation:
    """Helper function to return LemmatizeOperation

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    out : LemmatizeOperation
    """
    return LemmatizeOperation(nlp, name="lemmatize")


class TokenFilterOperation(SpacyBasedOperation):
    """Operation for filtering individual tokens.

    Spacy's token pattern matching is used for matching various
    tokens in the document. Any tokens matching the filter can
    either be discarded or kept while discarding the non matching ones.

    Parameters
    ----------
    patterns : List[List[Dict]]
        a list of patterns where each pattern is a List[Dict]. The patterns
        are passed to spacy's Token Matcher.
        see https://spacy.io/usage/rule-based-matching for more details
        on how to define patterns.

    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    keep_matching_tokens: bool
        if true then any non-matching tokens are discarded from the document (e.g. extracting only nouns)
        if false then any matching tokens are discarded (e.g. stopword removal)

    name : Optional[str]
        name of this operation

    Example
    -------
    >>> nlp = spacy.load("en_core_web_sm")
    >>> # define patterns to match [a, an, the] tokens
    >>> patterns = [
        [{"LOWER": "a"}],
        [{"LOWER": "an"}],
        [{"LOWER": "the"}]
    ]
    >>> # define the token filter operation to match the patterns and discard them
    >>> op = TokenFilterOperation(patterns=patterns, nlp=nlp, keep_matching_tokens=False)
    >>> ds = DataStream(["that is an orange"])
    >>> print(list(ds.apply(op))
    ["that is orange"]

    See https://spacy.io/usage/rule-based-matching#adding-patterns-attributes for more details
    on what token patterns can be used.

    Attributes
    ---------
    nlp : spacy.language.Language
        spacy's language model

    keep_matching_tokens : bool
        whether to discard the tokens matched by the filter from the document
        or to keep them

    patterns : List[List[Dict]]
        patterns to pass to spacy's Matcher

    name : str
        name of this operation

    """

    def __init__(
        self,
        patterns: List[List[Dict]],
        nlp: Optional[Language] = None,
        keep_matching_tokens=False,
        name: Optional[str] = "token_filter",
    ) -> None:
        super().__init__(nlp, name=name)
        self.keep_matching_tokens = keep_matching_tokens
        self.patterns = patterns
        self.matcher = self._get_matcher(self.nlp, self.patterns)

    def _get_matcher(self, nlp, patterns):
        matcher = Matcher(vocab=nlp.vocab, validate=True)

        for p in patterns:
            matcher.add("MATCHES", None, p)

        return matcher

    def _filter_tokens(self, matcher_output: Tuple[Doc, List[Tuple]]) -> Doc:
        doc, matches = matcher_output
        matching_token_ids = []
        for _, start, end in matches:
            for token in doc[start:end]:
                matching_token_ids.append(token.i)

        tokens_to_discard = matching_token_ids
        if self.keep_matching_tokens:
            tokens_to_discard = [t.i for t in doc if t.i not in matching_token_ids]

        return self.discard_tokens_from_doc(doc, tokens_to_discard)

    def run(self, ds: DataStream) -> DataStream:
        docs = self.get_docs(ds)
        match_results = self.matcher.pipe(docs, return_matches=True)
        new_docs = map(self._filter_tokens, match_results)
        for _, proc in self.nlp.pipeline:
            new_docs = proc.pipe(new_docs)
        return DataStream(
            new_docs, applied_ops=ds.applied_ops + [self], context=ds.context
        )

    def __getstate__(self):
        state = super().__getstate__()
        del state["matcher"]
        return state

    def __setstate__(self, state: dict):
        super().__setstate__(state)
        self.matcher = self._get_matcher(self.nlp, self.patterns)


def token_filter(
    patterns: List[List[Dict]],
    keep_matching_tokens,
    nlp: Optional[Language] = None,
    name: Optional[str] = "token_filter",
) -> TokenFilterOperation:
    """Helper function to create TokenFilterOperation

    Parameters
    ----------
    patterns : List[List[Dict]]
        a list of patterns where each pattern is a List[Dict]. The patterns
        are passed to spacy's Token Matcher.
        see https://spacy.io/usage/rule-based-matching for more details
        on how to define patterns.

    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    keep_matching_tokens: bool
        if true then any non-matching tokens are discarded from the document (e.g. extracting only nouns)
        if false then any matching tokens are discarded (e.g. stopword removal)

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    return TokenFilterOperation(
        patterns=patterns,
        nlp=nlp,
        keep_matching_tokens=keep_matching_tokens,
        name=name,
    )


def remove_stopwords(
    words: List[str],
    nlp: Optional[Language] = None,
    name: Optional[str] = "remove_stopwords",
) -> TokenFilterOperation:
    """TokenFilterOperation to remove stopwords

    Parameters
    ----------
    words : List[str]
        a list of words to remove from the text

    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    patterns = []
    for word in words:
        patterns.append([{"LOWER": word.lower()}])
    return TokenFilterOperation(
        patterns, nlp=nlp, keep_matching_tokens=False, name=name
    )


def remove_numbers(
    nlp: Optional[Language] = None, name: Optional[str] = "remove_numbers"
) -> TokenFilterOperation:
    """TokenFilterOperation to remove numbers

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    patterns = [[{"IS_DIGIT": True}]]
    return TokenFilterOperation(
        patterns, nlp=nlp, keep_matching_tokens=False, name=name
    )


def remove_links(
    nlp: Optional[Language] = None, name: Optional[str] = "remove_links"
) -> TokenFilterOperation:
    """TokenFilterOperation to remove hyperlinks

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    patterns = [[{"LIKE_URL": True}]]
    return TokenFilterOperation(
        patterns, nlp=nlp, keep_matching_tokens=False, name=name
    )


def remove_emails(
    nlp: Optional[Language] = None, name: Optional[str] = "remove_emails"
) -> TokenFilterOperation:
    """TokenFilterOperation to remove emails

    Parameters
    ----------
    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    patterns = [[{"LIKE_EMAIL": True}]]
    return TokenFilterOperation(
        patterns, nlp=nlp, keep_matching_tokens=False, name=name
    )


def remove_words_with_length_less_than(
    length: int,
    nlp: Optional[Language] = None,
    name: Optional[str] = "remove_words_with_length_less_than",
) -> TokenFilterOperation:
    """TokenFilterOperation to remove tokens that have fewer characters
    than specified

    Parameters
    ----------
    length : int
        atleast this many characters should be in the token, otherwise
        it is discarded

    nlp : Optional[spacy.language.Language]
        spacy's language model or None. If None then by default
        `en_core_web_sm` spacy model is loaded

    name : Optional[str]
        name of this operation

    Returns
    -------
    TokenFilterOperation
    """
    patterns = [[{"LENGTH": {"<": length}}]]
    return TokenFilterOperation(
        patterns, nlp=nlp, keep_matching_tokens=False, name=name,
    )



from jange.ops import base  # noqa:
from jange.ops import cluster  # noqa:
from jange.ops import dim  # noqa:
from jange.ops import text  # noqa:
from jange.ops import topic  # noqa:
from jange.ops import utils  # noqa:


__all__ = ["base", "cluster", "dim", "text", "topic", "utils"]



from typing import List, Optional, Union
import pandas as pd
from jange.base import DataStream

__all__ = ["DataFrameStream", "CSVDataStream", "from_csv", "from_df"]


class DataFrameStream(DataStream):
    """Represents a stream of data by iterating over the rows in a
    pandas DataFrame object.

    Parameters
    ----------
    df : pd.DataFrame
        pandas DataFrame object

    columns : Union[str, List[str]]
        a column name or a list of column names in the dataframe. The values
        from the given column(s) are used to create a stream. If a list is
        passed then each item in the stream will be a list of values for the given
        columns in that order.

    Example
    -------
    >>> df = pd.DataFrame([{"text": "text 1", "id": "1"}, {"text": "text 2", "id": "2"}])
    >>> ds = DataFrameStream(df=df, columns="text")
    >>> print(list(ds))
    >>> ["text 1", "text 2"]
    >>> ds = DataFrameStream(df=df, columns=["id", "text"])
    >>> print(list(ds))
    >>> [["1", "text 1"], ["2", "text 2"]]

    Attributes
    ----------
    df : pd.DataFrame
        a pandas DataFrame object
    columns : Union[str, list]
        a list of column names or a single column name. This value
        is used to select data from those columns only
    """

    def __init__(
        self, df, columns: Union[str, List[str]], context_column: Optional[str] = None
    ) -> None:
        super().__init__(applied_ops=None, items=["dummy"])
        self.context, self.items = self._get_items(df, columns, context_column)
        self.columns = columns
        self.context_column = context_column

    def _get_items(self, df, columns, context_column):
        contexts = []
        items = []
        for i, row in df.iterrows():
            context = row[context_column] if context_column else i
            contexts.append(context)
            if isinstance(columns, list):
                value = [row[c] for c in columns]
            else:
                value = row[columns]

            items.append(value)

        return contexts, items


class CSVDataStream(DataFrameStream):
    """Represents a stream of data by reading the contents from a csv file.
    pandas library is used to read the csv.

    Parameters
    ----------
    path : str
        path to the csv file to read. This parameter is passed directly to
        `pandas.read_csv` method

    columns : Union[str, List[str]]
        a column name or a list of column names in the csv file. The values
        from the given column(s) are used to create a stream. If a list is
        passed then each item in the stream will be a list of values for the given
        columns in that order.

    Example
    -------
    >>> ds = CSVDataStream(path="news_articles.csv", columns=["body", "title"])

    Attributes
    ----------
    df : pd.DataFrame
        a pandas DataFrame object created after reading the csv file
    columns : Union[str, list]
        a list of column names or a single column name. This value is used to
        select data from those columns only
    path : str
        path to the csv file
    """

    def __init__(
        self,
        path: str,
        columns: Union[str, List[str]],
        context_column: Optional[str] = None,
    ):
        super().__init__(
            pd.read_csv(path), columns=columns, context_column=context_column
        )
        self.path = path


def from_csv(path: str, columns: list) -> DataStream:
    return CSVDataStream(path=path, columns=columns)


def from_df(df, columns: list) -> DataStream:
    return DataFrameStream(df, columns)



from contextlib import contextmanager
from functools import lru_cache
from typing import List

import spacy
from jange.base import Operation, TrainableMixin


@contextmanager
def disable_training(ops: List[Operation]) -> List[Operation]:
    """Disables the "training mode" of operations so that these
    operations can be used for inference.

    Some operations like tfidf, kmeans, sgd etc. need to learn
    from the data and by default they are in "training mode"
    which will learn from the stream that is passed to them. Once
    these have been trained we want to use it in production so
    we need to disable the training.

    Example
    -------
    >>> with ops.utils.disable_training(stream.applied_ops) as new_ops:
    >>>     ops.utils.save(new_ops, path="./operations")

    Parameters
    ----------
    ops : List[Operation]
        a list of operations
    """
    original_mode = {}
    try:
        for i, op in enumerate(ops):
            if isinstance(op, TrainableMixin):
                original_mode[i] = op.should_train
                op.should_train = False

        yield ops
    finally:
        for i, op in enumerate(ops):
            if isinstance(op, TrainableMixin):
                op.should_train = original_mode[i]


def save(ops: List[Operation], path: str, replace: bool = False):
    pass


def load(path: str) -> List[Operation]:
    pass


@lru_cache(maxsize=32)
def cached_spacy_model(name_or_path: str):
    return spacy.load(name_or_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jange', 'jange.ops', 'jange.ops.text', 'jange.stream', 'jange.vis']

package_data = \
{'': ['*']}

install_requires = \
['cytoolz>=0.10.0,<0.11.0',
 'jsonschema>=3.2.0,<4.0.0',
 'pandas==0.24.2',
 'plotly>=4.8.2,<5.0.0',
 'psycopg2-binary>=2.8.3,<3.0.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'spacy>=2.2.0,<3.0.0',
 'sqlalchemy==1.3.1']

setup_kwargs = {
    'name': 'jange',
    'version': '0.1.1',
    'description': 'Easy NLP library for Python',
    'long_description': None,
    'author': 'Sanjaya Subedi',
    'author_email': 'jangedoo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from .clean import (
    CaseChangeOperation,
    lowercase,
    uppercase,
    LemmatizeOperation,
    lemmatize,
    ConvertToSpacyDocOperation,
    convert_to_spacy_doc,
    TokenFilterOperation,
    token_filter,
    remove_emails,
    remove_links,
    remove_numbers,
    remove_stopwords,
    remove_words_with_length_less_than,
)
from .vectorize import (
    SklearnBasedVectorizer,
    tfidf,
    count,
    DocumentEmbeddingOperation,
    doc_embedding,
)

__all__ = [
    "CaseChangeOperation",
    "lowercase",
    "uppercase",
    "LemmatizeOperation",
    "lemmatize",
    "ConvertToSpacyDocOperation",
    "convert_to_spacy_doc",
    "TokenFilterOperation",
    "token_filter",
    "remove_emails",
    "remove_links",
    "remove_numbers",
    "remove_stopwords",
    "remove_words_with_length_less_than",
    "SklearnBasedVectorizer",
    "tfidf",
    "count",
    "DocumentEmbeddingOperation",
    "doc_embedding",
]



import numpy as np
import pandas as pd
import plotly.express as px

from jange.stream import DataStream


def visualize(features, clusters):
    features = features.items if isinstance(features, DataStream) else features
    features = np.array(features)
    n_dim = features.shape[-1]
    if n_dim not in [2, 3]:
        raise ValueError(
            f"To visualize, the input matrix should be either 2D or 3D but got {n_dim}"
        )
    data = {}
    for axis_name, axis in zip(["x", "y", "z"], range(features.shape[-1])):
        data[axis_name] = features[:, axis]

    if isinstance(clusters, DataStream):
        context = clusters.context
        clusters = clusters.items
    else:
        context = range(len(clusters))

    data["cluster"] = clusters
    data["context"] = context
    df = pd.DataFrame(data)

    if n_dim == 2:
        fig = px.scatter(df, x="x", y="y", color="cluster", hover_data=["context"])
    else:
        fig = px.scatter_3d(
            df, x="x", y="y", z="z", color="cluster", hover_data=["context"]
        )
    return fig
